#!/usr/bin/python2.4
# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import service
from google.protobuf import service_reflection
from google.protobuf import descriptor_pb2



_REF = descriptor.Descriptor(
  name='Ref',
  full_name='Ref',
  filename='src/keas/pbstate/testclasses.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='_p_refid', full_name='Ref._p_refid', index=0,
      number=1, type=13, cpp_type=3, label=2,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_ADDRESSPB = descriptor.Descriptor(
  name='AddressPB',
  full_name='AddressPB',
  filename='src/keas/pbstate/testclasses.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='line1', full_name='AddressPB.line1', index=0,
      number=1, type=9, cpp_type=9, label=2,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='line2', full_name='AddressPB.line2', index=1,
      number=2, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='city', full_name='AddressPB.city', index=2,
      number=3, type=9, cpp_type=9, label=2,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='state', full_name='AddressPB.state', index=3,
      number=4, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='postal_code', full_name='AddressPB.postal_code', index=4,
      number=5, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='country', full_name='AddressPB.country', index=5,
      number=6, type=9, cpp_type=9, label=2,
      default_value=unicode("United States", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_CONTACTPB = descriptor.Descriptor(
  name='ContactPB',
  full_name='ContactPB',
  filename='src/keas/pbstate/testclasses.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='create_time', full_name='ContactPB.create_time', index=0,
      number=1, type=4, cpp_type=4, label=2,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='name', full_name='ContactPB.name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='address', full_name='ContactPB.address', index=2,
      number=3, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='guardians', full_name='ContactPB.guardians', index=3,
      number=4, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_CONTACTPB.fields_by_name['address'].message_type = _ADDRESSPB
_CONTACTPB.fields_by_name['guardians'].message_type = _REF

class Ref(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _REF

class AddressPB(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _ADDRESSPB

class ContactPB(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _CONTACTPB






class StateTuple(tuple):
    """Contains the persistent state of a ProtobufState object.

    Contains data (a byte string) and targets (a map of refid to object).
    This state class has an attribute, 'serial_format', that provides
    a hint on how the state ought to be serialized.
    """
    serial_format = 'protobuf'



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Provides the ProtobufState metaclass."""

from google.protobuf.descriptor import FieldDescriptor

from keas.pbstate.state import StateTuple


def _protobuf_property(name):
    """A property that delegates to a protobuf message"""
    def get(self):
        return getattr(self.protobuf, name)
    def set(self, value):
        setattr(self.protobuf, name, value)
    def delete(self):
        delattr(self.protobuf, name)
    return property(get, set, delete)


def _protobuf_mixin_property(container_getter, name):
    """A property that delegates to a protobuf sub-message"""
    def get_(self):
        return getattr(container_getter(self), name)
    def set_(self, value):
        setattr(container_getter(self), name, value)
    def del_(self):
        delattr(container_getter(self), name)
    return property(get_, set_, del_)


def _add_mixin(created_class, mixin_name):
    """Add the properties from a protobuf sub-message to a class"""
    main_desc = created_class.protobuf_type.DESCRIPTOR

    # traverse to the named descriptor
    descriptor = main_desc
    parts = mixin_name.split('.')
    for name in parts:
        # find the named field
        for f in descriptor.fields:
            if f.name == name:
                descriptor = f.message_type
                break
        else:
            raise AttributeError(
                "Field %r not defined for protobuf type %r" %
                (mixin_name, main_desc))

    container_getter = eval(
        'lambda self: self.protobuf.%s' % mixin_name)

    for field in descriptor.fields:
        setattr(created_class, field.name, _protobuf_mixin_property(
            container_getter, field.name))


class MessageRefidGatherer:
    """Returns the set of used refids from a message of a certain type."""

    def __init__(self, descriptor):
        self.attrs = {}  # {attr: gather callable}
        self.has_refs = False

        # Visit certain fields in the message.
        for field in descriptor.fields:

            # set up a 'gather' callable that iterates refids,
            if field.message_type is not None:
                gather = MessageRefidGatherer(field.message_type)
                if gather.has_refs:
                    self.has_refs = True
                else:
                    # don't bother visiting this part of the message
                    continue
            elif field.name == '_p_refid':
                self.has_refs = True
                def gather(value):
                    if value:
                        yield value
            else:
                # other scalars don't matter
                continue

            if field.label == FieldDescriptor.LABEL_REPEATED:
                def gather_all(container, gather=gather):
                    for obj in container:
                        for refid in gather(obj):
                            yield refid
                gather = gather_all

            self.attrs[field.name] = gather

    def __call__(self, container):
        for name, gather in self.attrs.iteritems():
            for refid in gather(getattr(container, name)):
                yield refid


class ProtobufReferences(object):
    """Gets or sets the target of references.

    A reference is a protobuf message with a _p_refid field.
    The target can be any kind of pickleable object, including derivatives
    of ZODB.Persistent.
    """
    __slots__ = ('_targets',)

    def __init__(self, targets):
        self._targets = targets  # {refid -> target}

    def get(self, ref_message, default=None):
        """Get a reference target"""
        refid = ref_message._p_refid
        if not refid:
            return default
        return self._targets[refid]

    def set(self, ref_message, target):
        """Set the target of a reference message"""
        targets = self._targets
        refid = id(target) % 0xffffffff
        if refid not in targets or targets[refid] is not target:
            while not refid or refid in targets:
                refid = (refid + 1) % 0xffffffff
            targets[refid] = target
        ref_message._p_refid = refid

    def delete(self, ref_message):
        """Unlink a target from a reference message"""
        # We can't actually remove the reference from the reference mapping
        # because something else might still have a reference.  __getstate__
        # will remove unused references.
        ref_message._p_refid = 0

    def _get_targets(self, obj, used):
        """Clean out unused reference targets, then return the target dict.

        Also raises an error if there are broken references.
        obj is the object that contains this references object.  used
        is the set of _p_refids in use by the protobuf.
        """
        targets = self._targets
        current = set(targets)
        broken = used.difference(current)
        if broken:
            raise KeyError("Object contains broken references: %s" % repr(obj))
        for key in current.difference(used):
            del targets[key]
        return targets


class StateClassMethods(object):
    """Contains methods that get copied into classes using ProtobufState"""

    def __getstate__(self):
        """Encode the entire state of the object in a ProtobufState."""
        # Clean up unused references and get the targets mapping.
        used = set(self._protobuf_find_refids(self.protobuf))
        targets = self.protobuf_refs._get_targets(self, used)
        if hasattr(self, '_p_changed'):
            # Reset the message's internal _cache_byte_size_dirty flag
            self.protobuf.ByteSize()
        # Return the state and all reference targets.
        return StateTuple((self.protobuf.SerializeToString(), targets))

    def __setstate__(self, state):
        """Set the state of the object.

        Accepts a StateTuple or any two item sequence containing
        data (a byte string) and targets (a map of refid to object).
        """
        data, targets = state
        self.protobuf_refs = ProtobufReferences(targets)
        self.protobuf = self.protobuf_type()
        self.protobuf.MergeFromString(data)
        if hasattr(self, '_p_changed'):
            self.protobuf._SetListener(PersistentChangeListener(self))
            # Reset the message's internal _cache_byte_size_dirty flag
            self.protobuf.ByteSize()


class PersistentChangeListener(object):
  """Propagates protobuf change notifications to a Persistent container.

  Implements the interface described by
  google.protobuf.internal.message_listener.MessageListener.
  """
  __slots__ = ('obj',)

  def __init__(self, obj):
      self.obj = obj

  def TransitionToNonempty(self):
      self.obj._p_changed = True

  def ByteSizeDirty(self):
      self.obj._p_changed = True


class ProtobufState(type):
    """Metaclass for classes using a protobuf for state and serialization."""

    def __new__(metaclass, name, bases, dct):
        """Set up a new class."""

        # Arrange for class instances to always have the 'protobuf' and
        # 'protobuf_refs' instance attributes.  This is done by creating
        # or overriding the __new__() method of the new class.
        def __new__(subclass, *args, **kw):
            # subclass is either created_class or a subclass of created_class.
            super_new = super(created_class, subclass).__new__
            instance = super_new(subclass, *args, **kw)
            instance.protobuf_refs = ProtobufReferences({})
            instance.protobuf = subclass.protobuf_type()
            if hasattr(instance, '_p_changed'):
                instance.protobuf._SetListener(
                    PersistentChangeListener(instance))
            return instance
        dct['__new__'] = __new__

        # Limit instance attributes to avoid partial serialization.
        # Note that classes can still store temporary state by
        # declaring other attribute names in the class' __slots__
        # attribute.
        dct['__slots__'] = dct.get('__slots__', ()) + (
            'protobuf', 'protobuf_refs')

        created_class = type.__new__(metaclass, name, bases, dct)
        if not hasattr(created_class, 'protobuf_type'):
            raise TypeError("Class %s.%s needs a protobuf_type attribute"
                % (created_class.__module__, created_class.__name__))
        descriptor = created_class.protobuf_type.DESCRIPTOR

        # Copy methods into the class
        for method_name in ('__getstate__', '__setstate__'):
            setattr(created_class, method_name,
                StateClassMethods.__dict__[method_name])

        # Set up the refid gatherer
        created_class._protobuf_find_refids = MessageRefidGatherer(descriptor)

        # Create properties that delegate storage to the protobuf
        for field in descriptor.fields:
            setattr(created_class, field.name,
                _protobuf_property(field.name))

        # Create properties that delegate to mixed in attributes
        for mixin_name in dct.get('protobuf_mixins', ()):
            _add_mixin(created_class, mixin_name)

        return created_class



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""An object database foundation based on Google's Protocol Buffers"""

import os
from setuptools import setup

VERSION='0.1'

def read_file(*path):
    base_dir = os.path.dirname(__file__)
    file_path = (base_dir, ) + tuple(path)
    return file(os.path.join(*file_path)).read()

setup(
    name='keas.pbstate',
    version=VERSION,
    author='Shane Hathaway and the Zope Community',
    author_email='zope-dev@zope.org',
    description=__doc__,
    license='ZPL 2.1',

    package_dir={'': 'src'},
    packages=['keas.pbstate'],
    namespace_packages=['keas'],
    install_requires=[
        'setuptools',
        'protobuf',
        ],
    long_description = read_file('src', 'keas', 'pbstate', 'README.txt'),
)
